from typing import List

from database_mysql_local.generic_crud import GenericCRUD
from database_mysql_local.point import Point
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .county_ml import CountyMl
from .location_local_constants import LocationLocalConstants
from .util import LocationsUtil

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class County(GenericCRUD):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init County")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.COUNTY_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.COUNTY_ML_VIEW_NAME,
            default_column_name=LocationLocalConstants.COUNTY_ID_COLUMN_NAME,
            is_test_data=is_test_data
        )

        self.county_ml = None

        logger.end("end init County")

    def insert(self, *,  # noqa
               county: str, state_id: int, lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE,
               title_approved: bool = False, coordinate: Point = None,
               group_id: int = None) -> int or None:
        logger.start("start insert county",
                     object={'coordinate': coordinate, 'county': county,
                             'lang_code': lang_code, 'state_id': state_id,
                             'title_approved': title_approved})
        is_valid = LocationsUtil.validate_insert_args(
            name=county, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="County was not inserted because no county name was provided")
            return None
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        county_dict = {'coordinate': coordinate,
                       'group_id': group_id,
                       'state_id': state_id}

        county_id = super().insert(data_dict=county_dict)

        self.county_ml = self.county_ml or CountyMl()
        county_ml_id = self.county_ml.insert(county_id=county_id,
                                             county=county,
                                             lang_code=lang_code,
                                             title_approved=title_approved)

        logger.end("end insert county",
                   object={'county_id': county_id,
                           'county_ml_id': county_ml_id})
        return county_id

    # TODO read or get? - What is our standard?
    # TODO get_location_dict_by_location_id(...) - We might have Location in the future.
    def read_by_location(self, location_id: int) -> dict:
        logger.start("start read location", object={'location_id': location_id})
        result = super().select_one_dict_by_column_and_value(
            column_value=location_id,
            select_clause_value=LocationLocalConstants.COUNTY_TABLE_COLUMNS)

        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_dict=result)

        logger.end("end read location", object={"result": result})
        return result

    def read_by_location_and_state(self, location_id: int, state_id: int) -> List[dict]:
        logger.start("start read location", object={'location_id': location_id})
        result = self.select_multi_dict_by_where(
            where=f"{LocationLocalConstants.LOCATION_ID_COLUMN_NAME}={location_id} AND "
                  f"{LocationLocalConstants.STATE_ID_COLUMN_NAME}={state_id}",
            select_clause_value=LocationLocalConstants.COUNTY_TABLE_COLUMNS)

        result = [LocationsUtil.extract_coordinates_and_replace_by_point(data_dict=county) for county in result]

        logger.end("end read location", object={"result": result})
        return result

    def get_county_id_by_county_name_state_id(
            self, *, county_name: str, state_id: int = None,
            lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int or None:
        logger.start("start get_county_id_by_county_name_state_id",
                     object={'county_name': county_name, 'state_id': state_id})
        if county_name is None:
            logger.end(log_message="end get_county_id_by_county_name_state_id",
                       object={'county_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(county_name)
        where_clause = f"title='{county_name}' AND lang_code='{lang_code.value}'"
        if state_id is not None:
            where_clause += f" AND state_id={state_id}"

        self.county_ml = self.county_ml or CountyMl()
        county_id_dict = self.county_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.COUNTY_ID_COLUMN_NAME,
            where=where_clause,
            order_by="county_id DESC")
        # TODO logger.info(country_id_dict)
        # TODO Error handing, what happen if we don't have location, we should catch and raise specific exception
        county_id = county_id_dict.get(
            LocationLocalConstants.COUNTY_ID_COLUMN_NAME)

        logger.end("end get_county_id_by_county_name_state_id", object={'county_id': county_id})
        return county_id



from typing import List

from database_mysql_local.generic_crud import GenericCRUD
from database_mysql_local.point import Point
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .city_ml import CityMl
from .location_local_constants import LocationLocalConstants
from .util import LocationsUtil

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


# TODO: migrate all classes to meta logger
class City(GenericCRUD):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init City")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.CITY_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.CITY_ML_VIEW_NAME,
            default_column_name=LocationLocalConstants.CITY_ID_COLUMN_NAME,
            is_test_data=is_test_data)
        self.city_ml = None

        logger.end("end init City")

    def insert(self, *,  # noqa
               city: str, state_id: int, lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE,
               title_approved: bool = False, coordinate: Point = None,
               group_id: int = None, phonecode: int = None,
               is_main: int = None) -> int or None:
        """A city can have multiple state_id
        https://www.cheapflights.com/news/seeing-double-us-cities-nice-used-name-twice"""
        logger.start("start insert city",
                     object={'coordinate': coordinate, 'city': city,
                             'lang_code': lang_code,
                             'state_id': state_id,
                             'title_approved': title_approved,
                             'is_main': is_main})
        is_valid = LocationsUtil.validate_insert_args(
            name=city, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="City was not inserted because no city name was provided")
            return None
        lang_code = lang_code or LangCode.detect_lang_code(city)
        city_dict = {
            key: value for key, value in {
                'coordinate': coordinate,
                'name': city,
                'state_id': state_id,
                'phonecode': phonecode,
                'group_id': group_id
            }.items() if value is not None
        }

        city_id = super().insert(data_dict=city_dict)

        self.city_ml = self.city_ml or CityMl()
        city_ml_id = self.city_ml.insert(city_id=city_id,
                                         city=city,
                                         lang_code=lang_code,
                                         title_approved=title_approved,
                                         is_main=is_main)

        logger.end("end insert city",
                   object={'city_id': city_id,
                           'city_ml_id': city_ml_id})
        return city_id

    def read_by_location(self, location_id: int) -> dict:
        """Read city by location_id"""
        logger.start("start read city", object={'location_id': location_id})
        result = super().select_one_dict_by_column_and_value(
            column_value=location_id,
            select_clause_value=LocationLocalConstants.CITY_TABLE_COLUMNS)
        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_dict=result)
        logger.end("end read location", object={"result": result})
        return result

    def read_by_location_and_state(self, location_id: int, state_id: int) -> List[dict]:
        logger.start("start read city", object={'location_id': location_id, 'state_id': state_id})
        result = self.select_multi_dict_by_where(
            where=f"{LocationLocalConstants.CITY_ID_COLUMN_NAME}={location_id} AND "
                  f"{LocationLocalConstants.STATE_ID_COLUMN_NAME}={state_id}",
            select_clause_value=LocationLocalConstants.CITY_TABLE_COLUMNS)
        result = [LocationsUtil.extract_coordinates_and_replace_by_point(data_dict=city) for city in result]
        logger.end("end read location", object={"result": result})
        return result

    def get_city_id_by_city_name(self, city_name: str,
                                 lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int or None:
        logger.start("start get_city_id_by_city_name_state_id",
                     object={'city_name': city_name})
        if city_name is None:
            logger.end(log_message="end get_city_id_by_city_name_state_id",
                       object={'city_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(city_name)
        where_clause = "title=%s AND lang_code=%s"
        params = (city_name, lang_code.value)

        self.city_ml = self.city_ml or CityMl()
        city_id_dict = self.city_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.CITY_ID_COLUMN_NAME,
            where=where_clause,
            params=params,
            order_by="city_id DESC")
        city_id = city_id_dict.get(LocationLocalConstants.CITY_ID_COLUMN_NAME)

        logger.end("end get_city_id_by_city_name_state_id", object={'city_id': city_id})
        return city_id



from database_mysql_local.generic_crud_ml import GenericCRUDML
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants

logger = Logger.create_logger(object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class CountryMl(GenericCRUDML):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init CountryMl")
        GenericCRUDML.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.COUNTRY_TABLE_NAME,
            default_column_name=LocationLocalConstants.COUNTRY_ML_ID_COLUMN_NAME,
            default_view_table_name=LocationLocalConstants.COUNTRY_ML_VIEW_NAME,
            default_ml_view_table_name=LocationLocalConstants.COUNTRY_ML_VIEW_NAME,
            is_test_data=is_test_data)
        logger.end("end init CountryMl")

    def insert(self, *,  # noqa
               country_id: int, country: str, title_approved: bool = False,
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int:
        logger.start("start insert country_ml",
                     object={'country_id': country_id, 'country': country,
                             'lang_code': lang_code,
                             'title_approved': title_approved})
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(country)
        country_ml_dict = {
            'country_id': country_id,
            'lang_code': lang_code.value,
            'title': country,
            'title_approved': title_approved
        }
        country_ml_id = super().insert(
            table_name=LocationLocalConstants.COUNTRY_ML_TABLE_NAME,
            data_dict=country_ml_dict)
        logger.end("end insert country_ml",
                   object={'country_ml_id': country_ml_id})

        return country_ml_id



from typing import Dict

from database_mysql_local.generic_crud import GenericCRUD
from database_mysql_local.point import Point
from language_remote.lang_code import LangCode
from logger_local.LoggerComponentEnum import LoggerComponentEnum
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .city import City
from .country import CountriesLocal
from .county import County
from .location_local_constants import LocationLocalConstants
from .neighborhood import Neighborhood
from .region import Region
from .state import State
from .util import LocationsUtil

DUMMY_POINT = Point(0, 0)
LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID = LocationLocalConstants.LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID
LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME = LocationLocalConstants.LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME

object_to_insert = {
    'component_id': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID,
    'component_name': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': 'tal.g@circ.zone'
}

logger = Logger.create_logger(object=object_to_insert)

user_context = UserContext()


# TODO Create LocationsLocal.get_country_id_by_location_id() and call it
# from importer-local-python-package

# TODO: add location ml using crud ml
class LocationsLocal(GenericCRUD):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init LocationLocal")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.LOCATION_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.LOCATION_VIEW_NAME,
            default_column_name=LocationLocalConstants.LOCATION_ID_COLUMN_NAME,
            is_test_data=is_test_data)

        # Initiated with None for performance reasons
        self.city_instance = None
        self.country_instance = None
        self.county_instance = None
        self.neighborhood_instance = None
        self.region_instance = None
        self.state_instance = None

        logger.end("end init LocationLocal")

    def get_location_dict(self, *, neighborhood_name: str, county_name: str, region_name: str, state_name: str,
                          country_name: str, city_name: str, lang_codes_dict: dict = None) -> dict:
        # TODO Fix the algorithm as the current will not work in all cases. i.e. same neighborhood name in different places
        logger.start("start get location ids",
                     object={'neighborhood_name': neighborhood_name,
                             'county_name': county_name,
                             'region_name': region_name,
                             'state_name': state_name,
                             'country_name': country_name,
                             'city_name': city_name})
        lang_codes_dict = lang_codes_dict or {}
        self.city_instance = self.city_instance or City()
        self.country_instance = self.country_instance or CountriesLocal()
        self.county_instance = self.county_instance or County()
        self.neighborhood_instance = self.neighborhood_instance or Neighborhood()
        self.region_instance = self.region_instance or Region()
        self.state_instance = self.state_instance or State()

        neighborhood_id = self.neighborhood_instance.get_neighborhood_id_by_neighborhood_name(
            neighborhood_name=neighborhood_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "neighborhood"))
        county_id = self.county_instance.get_county_id_by_county_name_state_id(
            county_name=county_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "county"))
        region_id = self.region_instance.get_region_id_by_region_name(
            region_name=region_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "region"))
        state_id = self.state_instance.get_state_id_by_state_name(
            state_name=state_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "state"))
        country_id = self.country_instance.get_country_id_by_country_name(
            country_name=country_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "country"))
        city_id = self.city_instance.get_city_id_by_city_name(
            city_name=city_name,
            lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "city"))
        location_dict = {"neighborhood_id": neighborhood_id, "county_id": county_id, "region_id": region_id,
                         "state_id": state_id, "country_id": country_id, "city_id": city_id}
        logger.end(object=location_dict)
        return location_dict

    def insert(self, *,  # noqa
               data: Dict[str, any],
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE,
               is_approved: bool = False,
               is_test_data: bool = False,
               lang_codes_dict: dict = None,
               new_country_data: Dict[str, any] = None) -> int or None:

        logger.start("start insert location",
                     object={'data': data, 'lang_code': lang_code,
                             'is_approved': is_approved, 'is_test_data': is_test_data,
                             'new_country_data': new_country_data})
        if not data:
            logger.warning(log_message="Location was not inserted because no data was provided")
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        lang_codes_dict = lang_codes_dict or {}
        lang_codes_dict["default"] = lang_code
        location_dict = self._check_details_and_insert_if_not_exist(
            data=data, lang_codes_dict=lang_codes_dict,
            is_approved=is_approved, new_country_data=new_country_data)

        location_dict = {
            key: value for key, value in {
                'coordinate': data.get("coordinate"),
                'address_local_language': data.get("address_local_language"),
                'address_english': data.get("address_english"),
                'neighborhood_id': location_dict.get("neighborhood_id"),
                'county_id': location_dict.get("county_id"),
                'region_id': location_dict.get("region_id"),
                'state_id': location_dict.get("state_id"),
                'country_id': location_dict.get("country_id"),
                'city_id': location_dict.get("city_id"),
                'postal_code': data.get("postal_code"),
                'plus_code': data.get("plus_code"),
                'is_approved': is_approved,
                'is_test_data': is_test_data
            }.items() if value is not None
        }
        # We want to allow multiple locations for the same coordinates
        # We could do that if we had the location name
        # location_id = self.select_one_value_by_id(select_clause_value="location_id",
        #                                           column_name="coordinate", column_value=data.get("coordinate"))
        if data.get("coordinate") == DUMMY_POINT:  # used for tests
            location_id = 0
        else:
            location_id = super().insert(data_dict=location_dict)

        logger.end("end_insert location", object={'location_id': location_id})
        return location_id

    def update(self, *, location_id: int, data: Dict[str, any],
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE, is_approved: bool = False,
               lang_codes_dict: dict = None):

        logger.start("start update location",
                     object={'location_id': location_id, 'data': data,
                             'lang_code': lang_code,
                             'is_approved': is_approved})
        lang_codes_dict = lang_codes_dict or {}
        lang_codes_dict["default"] = lang_code
        location_info = self._check_details_and_insert_if_not_exist(
            data=data,
            lang_codes_dict=lang_codes_dict,
            is_approved=is_approved)

        updated_location_dict = {
            key: value for key, value in {
                'coordinate': data.get('coordinate'),
                'address_local_language': data.get(
                    "address_local_language"),
                'address_english': data.get("address_english"),
                'neighborhood_id': location_info.get("neighborhood_id"),
                'county_id': location_info.get("county_id"),
                'region_id': location_info.get("region_id"),
                'state_id': location_info.get("state_id"),
                'country_id': location_info.get("country_id"),
                'city_id': location_info.get("city_id"),
                'postal_code': data.get("postal_code"),
                'plus_code': data.get("plus_code"),
                'is_approved': is_approved
            }.items() if value is not None
        }
        super().update_by_column_and_value(
            column_value=location_id,
            data_dict=updated_location_dict
        )

        logger.end("end update location")

    def read(self, location_id: int) -> dict:
        logger.start("start read location",
                     object={'location_id': location_id})
        result = super().select_one_dict_by_column_and_value(
            column_value=location_id,
            select_clause_value=LocationLocalConstants.LOCATION_TABLE_COLUMNS)

        result = LocationsUtil.extract_coordinates_and_replace_by_point(data_dict=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    def delete(self, location_id: int):
        logger.start("start delete location by id",
                     object={'location_id': location_id})
        super().delete_by_column_and_value(column_value=location_id)

        logger.end("end delete location by id")

    def _check_details_and_insert_if_not_exist(
            self, *, data: dict,
            lang_codes_dict: dict = None, is_approved: bool = False,
            new_country_data: Dict[str, any] = None) -> dict or None:
        logger.start("start _check_details_and_insert_if_not_exist",
                     object={'data': data,
                             'new_country_data': new_country_data})
        coordinate = data.get("coordinate")
        neighborhood_name = data.get("neighborhood")
        county_name = data.get("county")
        region_name = data.get("region")
        state_name = data.get("state")
        country_name = data.get("country")
        city_name = data.get("city")

        location_info = self.get_location_dict(neighborhood_name=neighborhood_name, county_name=county_name,
                                               region_name=region_name, state_name=state_name,
                                               country_name=country_name,
                                               city_name=city_name, lang_codes_dict=lang_codes_dict)
        if location_info is None:
            return None

        if location_info.get("neighborhood_id") is None and neighborhood_name is not None:
            self.neighborhood_instance = self.neighborhood_instance or Neighborhood()
            location_info["neighborhood_id"] = self.neighborhood_instance.insert(
                coordinate=coordinate,
                neighborhood=neighborhood_name,
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "neighborhood"),
                title_approved=is_approved)

        if location_info.get("county_id") is None and country_name is not None:
            self.county_instance = self.county_instance or County()
            location_info["county_id"] = self.county_instance.insert(
                coordinate=coordinate,
                county=county_name,
                state_id=location_info["state_id"],
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "county"),
                title_approved=is_approved)

        if location_info.get("region_id") is None and region_name is not None:
            self.region_instance = self.region_instance or Region()
            location_info["region_id"] = self.region_instance.insert(
                coordinate=coordinate,
                region=region_name,
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "region"),
                title_approved=is_approved)

        if location_info.get("state_id") is None and state_name is not None:
            self.state_instance = self.state_instance or State()
            location_info["state_id"] = self.state_instance.insert(
                coordinate=coordinate,
                state=state_name,
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "state"),
                state_name_approved=is_approved)

        if location_info.get("country_id") is None and country_name is not None:
            self.country_instance = self.country_instance or CountriesLocal()
            location_info["country_id"] = self.country_instance.insert(
                coordinate=coordinate,
                country=country_name,
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "country"),
                title_approved=is_approved,
                new_country_data=new_country_data)
        if location_info.get("city_id") is None and city_name is not None:
            self.city_instance = self.city_instance or City()
            location_info["city_id"] = self.city_instance.insert(
                coordinate=coordinate,
                city=city_name,
                state_id=location_info["state_id"],
                lang_code=LocationsUtil.get_lang_code(lang_codes_dict, "city"),
                title_approved=is_approved)
        logger.end("end _check_details_and_insert_if_not_exist",
                   object=location_info)
        return location_info

    def get_test_location_id(self):
        logger.start("start get_test_location_id")
        test_point = Point(0, 0)
        test_location_id = self.get_test_entity_id(
            entity_name="location",
            insert_function=self.insert,
            insert_kwargs={"data": {"coordinate": test_point}},
            view_name=LocationLocalConstants.LOCATION_VIEW_NAME)
        logger.end("end get_test_location_id", object={'test_location_id': test_location_id})
        return test_location_id



from database_mysql_local.generic_crud_ml import GenericCRUDML
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class RegionMl(GenericCRUDML):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init RegionMl")
        GenericCRUDML.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.REGION_TABLE_NAME,
            default_column_name=LocationLocalConstants.REGION_ML_ID_COLUMN_NAME,
            default_view_table_name=LocationLocalConstants.REGION_ML_VIEW_NAME,
            default_ml_view_table_name=LocationLocalConstants.REGION_ML_VIEW_NAME,
            is_test_data=is_test_data)
        logger.end("end init RegionMl")

    def insert(self, *,  # noqa
               region_id: int, region: str, title_approved: bool = False,
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int:
        logger.start("start insert region_ml",
                     object={'region_id': region_id,
                             'region': region,
                             'lang_code': lang_code,
                             'title_approved': title_approved})
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(region)
        region_ml_dict = {
            key: value for key, value in {
                'region_id': region_id,
                'lang_code': lang_code.value,
                'title': region,
                'title_approved': title_approved
            }.items() if value is not None
        }
        region_ml_id = super().insert(
            table_name=LocationLocalConstants.REGION_ML_TABLE_NAME,
            data_dict=region_ml_dict)
        logger.end("end insert region_ml",
                   object={'region_ml_id': region_ml_id})

        return region_ml_id



from database_mysql_local.generic_crud_ml import GenericCRUDML
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants

logger = Logger.create_logger(object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class CountyMl(GenericCRUDML):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init CountyMl")
        super().__init__(
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.COUNTY_ML_TABLE_NAME,
            default_column_name=LocationLocalConstants.COUNTY_ML_ID_COLUMN_NAME,
            default_view_table_name=LocationLocalConstants.COUNTY_ML_VIEW_NAME,
            default_ml_view_table_name=LocationLocalConstants.COUNTY_ML_VIEW_NAME,
            is_test_data=is_test_data)
        logger.end("end init CountyMl")

    def insert(  # noqa
            self, *, county_id: int, county: str, title_approved: bool = False,
            lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int:
        logger.start("start insert county_ml",
                     object={'county_id': county_id, 'county': county,
                             'lang_code': lang_code,
                             'title_approved': title_approved})
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(county)
        county_ml_dict = {
            key: value for key, value in {
                'county_id': county_id,
                'lang_code': lang_code.value,
                'title': county,
                'title_approved': title_approved
            }.items() if value is not None
        }
        county_ml_id = super().insert(
            table_name=LocationLocalConstants.COUNTY_ML_TABLE_NAME,
            data_dict=county_ml_dict)
        logger.end("end insert county_ml",
                   object={'county_ml_id': county_ml_id})

        return county_ml_id



from database_mysql_local.generic_crud_ml import GenericCRUDML
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class StateMl(GenericCRUDML):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init StateMl")
        GenericCRUDML.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.STATE_TABLE_NAME,
            default_column_name=LocationLocalConstants.STATE_ML_ID_COLUMN_NAME,
            default_view_table_name=LocationLocalConstants.STATE_ML_VIEW_NAME,
            default_ml_view_table_name=LocationLocalConstants.STATE_ML_VIEW_NAME,
            is_test_data=is_test_data)
        logger.end("end init StateMl")

    def insert(self, *,  # noqa
               state_id: int, state: str, title_approved: bool = False,
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int:
        logger.start("start insert state_ml",
                     object={'state_id': state_id,
                             'state': state,
                             'lang_code': lang_code,
                             'title_approved': title_approved})
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(state)
        state_ml_dict = {
            key: value for key, value in {
                'state_id': state_id,
                'lang_code': lang_code.value,
                'state_name': state,
                'state_name_approved': title_approved
            }.items() if value is not None
        }
        state_ml_id = super().insert(
            table_name=LocationLocalConstants.STATE_ML_TABLE_NAME,
            data_dict=state_ml_dict)
        logger.end("end insert state_ml",
                   object={'state_ml_id': state_ml_id})

        return state_ml_id



from database_mysql_local.generic_crud_ml import GenericCRUDML
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class CityMl(GenericCRUDML):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init CityMl")
        GenericCRUDML.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.CITY_TABLE_NAME,
            default_column_name=LocationLocalConstants.CITY_ML_ID_COLUMN_NAME,
            default_view_table_name=LocationLocalConstants.CITY_ML_VIEW_NAME,
            default_ml_view_table_name=LocationLocalConstants.CITY_ML_VIEW_NAME,
            is_test_data=is_test_data)
        logger.end("end init CityMl")

    def insert(self, *,  # noqa
               city_id: int, city: str, title_approved: bool = False, is_main: int = None,
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int:
        logger.start("start insert city_ml",
                     object={'city_id': city_id, 'city': city,
                             'lang_code': lang_code,
                             'title_approved': title_approved,
                             'is_main': is_main})
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(city)
        city_ml_dict = {
            key: value for key, value in {
                'city_id': city_id,
                'lang_code': lang_code.value,
                'is_main': is_main,
                'title': city,
                'title_approved': title_approved
            }.items() if value is not None
        }
        # TODO Shall we use CrudMl in all those case in this file?
        city_ml_id = super().insert(
            table_name=LocationLocalConstants.CITY_ML_TABLE_NAME,
            data_dict=city_ml_dict)
        logger.end("end insert city_ml",
                   object={'city_ml_id': city_ml_id})

        return city_ml_id



from typing import Dict

from database_mysql_local.generic_crud import GenericCRUD
from database_mysql_local.point import Point
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from opencage.geocoder import OpenCageGeocode
from phonenumbers import geocoder, parse
from python_sdk_remote.utilities import our_get_env
from user_context_remote.user_context import UserContext

from .country_ml import CountryMl
from .location_local_constants import LocationLocalConstants
from .util import LocationsUtil

env_key = "OPENCAGE_KEY"

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


# TODO Can we use GenericMl
class CountriesLocal(GenericCRUD):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init Country")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.COUNTRY_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.COUNTRY_ML_VIEW_NAME,
            default_column_name=LocationLocalConstants.COUNTRY_ID_COLUMN_NAME,
            is_test_data=is_test_data)
        self.country_ml = None
        self.geocoder = OpenCageGeocode(our_get_env("OPENCAGE_KEY"))

        logger.end("end init Country")

    def insert(self, *, country: str,  # noqa
               lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE,
               title_approved: bool = False,
               new_country_data: Dict[str, any] = None,
               coordinate: Point = None, ignore_duplicate: bool = False) -> int or None:
        logger.start("start insert country",
                     object={'coordinate': coordinate, 'country': country,
                             'lang_code': lang_code,
                             'title_approved': title_approved,
                             'new_country_data': new_country_data})
        is_valid = LocationsUtil.validate_insert_args(
            name=country, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="Country was not inserted because no country name was provided")
            return None
        lang_code = lang_code or LangCode.detect_lang_code(country)
        new_country_data = new_country_data or {}
        try:
            country_dict = {
                key: value for key, value in {
                    'coordinate': coordinate,
                    'iso': new_country_data.get("iso"),
                    'name': country,
                    'nicename': new_country_data.get("nicename"),
                    'iso3': new_country_data.get("iso3"),
                    'numcode': new_country_data.get("numcode"),
                    'phonecode': new_country_data.get("phonecode")
                }.items() if value is not None
            }
            # TODO Can we use GenericMl
            country_id = super().insert(data_dict=country_dict, ignore_duplicate=ignore_duplicate)

        except Exception as e:
            logger.exception("error in insert country")
            logger.end()
            raise e
        try:
            self.country_ml = self.country_ml or CountryMl()
            country_ml_id = self.country_ml.insert(
                country=country,
                country_id=country_id,
                lang_code=lang_code,
                title_approved=title_approved)
        # TODO Replace e with exception everywhere
        except Exception as e:
            logger.exception("error in insert country")
            logger.end()
            raise e
        logger.end("end insert country",
                   object={'country_id': country_id,
                           'country_ml_id': country_ml_id})
        return country_id

    # TODO: the read function is duplicated in all classes
    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = super().select_one_dict_by_column_and_value(
            column_value=location_id,
            select_clause_value=LocationLocalConstants.COUNTRY_TABLE_COLUMNS)

        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_dict=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    def get_country_id_by_country_name(self, country_name: str,
                                       lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int or None:
        logger.start("start get_country_id_by_country_name",
                     object={'country_name': country_name})
        if country_name is None:
            logger.end(log_message="end get_country_id_by_country_name",
                       object={'country_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(country_name)
        where_clause = f"title='{country_name}' AND lang_code='{lang_code.value}'"

        self.country_ml = self.country_ml or CountryMl()
        country_id_dict = self.country_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.COUNTRY_ID_COLUMN_NAME,
            where=where_clause,
            order_by="country_id DESC")
        country_id = country_id_dict.get(LocationLocalConstants.COUNTRY_ID_COLUMN_NAME)

        logger.end("end get_country_id_by_country_name",
                   object={'country_id': country_id})
        return country_id

    def get_country_name(self, location: str) -> str:
        # Create a geocoder instance
        logger.start("start get_country_name",
                     object={'location': location})

        # Define the city or state
        # TODO Please create new class GeoCode and GeoCodeOpenCage Class which interit from GeoCode Class.
        # Please move those calls to the GeoCodeOpenCage.geocode(location).
        # TODO Please add calls to ApiManagement Indirect from GeoCode class

        # Use geocoding to get the location details
        results = self.geocoder.geocode(location)

        if results and len(results) > 0:
            first_result = results[0]
            components = first_result['components']

            # Extract the country from components
            country_name = components.get('country', '')
            if not country_name:
                # If country is not found, check for country_code
                # as an alternative
                country_name = components.get('country_code', '')
        else:
            country_name = None
            logger.error("country didnt  found for %s." % location)
        logger.end("end get_country_name",
                   object={'country_name': country_name})
        return country_name

    @staticmethod
    def get_country_name_by_phone_number(phone_number: str) -> str:
        logger.start("start get_country_name_by_phone_number",
                     object={'phone_number': phone_number})
        # Parse the phone number
        parsed_number = parse(phone_number)
        # Get the country from the phone number
        country_name = geocoder.country_name_for_number(parsed_number, 'en')
        if country_name == '' and phone_number.startswith('+1'):
            # Default to United States if the country is not found and the phone number starts with +1
            country_name = 'United States'
        if country_name == '' and phone_number.startswith('+972'):
            # Default to Israel if the country is not found and the phone number starts with +972
            country_name = 'Israel'
        logger.end("Successfully retrieved country name",
                   object={'phone_number': phone_number, 'country_name': country_name})
        return country_name.upper()

    @staticmethod
    def get_country_name_by_email_address(email_address: str) -> str:
        logger.start("start get_country_name_by_email_address",
                     object={'email_address': email_address})
        # Extract domain from the email address
        domain_parts = email_address.split('@')[-1].split('.')
        # Create domain suffixes that might include second-level domains
        domain_suffixes = [
            '.'.join(domain_parts[-(i + 1):]) for i in range(min(2, len(domain_parts) - 1))
        ]
        # Check each possible domain suffix until a match is found
        for suffix in domain_suffixes:
            if suffix in LocationLocalConstants.DOMAIN_TO_COUNTRY:
                country_name = LocationLocalConstants.DOMAIN_TO_COUNTRY.get(suffix)
                logger.end("Successfully retrieved country name",
                           object={'email_address': email_address, 'country_name': country_name})
                return country_name.upper()
        logger.end("Country name not found",
                   object={'email_address': email_address})


# TODO Create Country class which store one country
class Country(CountriesLocal):
    pass  # backward compatibility



from database_mysql_local.generic_crud import GenericCRUD
from database_mysql_local.point import Point
from language_remote.lang_code import LangCode
from logger_local.LoggerLocal import Logger
from user_context_remote.user_context import UserContext

from .location_local_constants import LocationLocalConstants
from .state_ml import StateMl
from .util import LocationsUtil

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class State(GenericCRUD):
    def __init__(self, is_test_data: bool = False):
        logger.start("start init State")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.STATE_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.STATE_ML_VIEW_NAME,
            default_column_name=LocationLocalConstants.STATE_ID_COLUMN_NAME,
            is_test_data=is_test_data)

        self.state_ml = None
        logger.end("End init State")

    def insert(self, *,  # noqa
               coordinate: Point,
               state: str, lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE,
               state_name_approved: bool = False,
               group_id: int = None) -> int or None:
        logger.start("start insert state",
                     object={'coordinate': coordinate, 'state': state,
                             'lang_code': lang_code,
                             'state_name_approved': state_name_approved,
                             'group_id': group_id})
        is_valid = LocationsUtil.validate_insert_args(
            name=state, lang_code=lang_code, title_approved=state_name_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="State was not inserted because no state name was provided")
            return None
        lang_code = lang_code or LangCode.detect_lang_code(state)
        state_dict = {
            key: value for key, value in {
                'coordinate': coordinate,
                'group_id': group_id
            }.items() if value is not None
        }

        state_id = super().insert(data_dict=state_dict)

        self.state_ml = self.state_ml or StateMl()
        state_ml_id = self.state_ml.insert(state_id=state_id, state=state,
                                           lang_code=lang_code,
                                           title_approved=state_name_approved)

        logger.end("end insert state",
                   object={'state_id': state_id,
                           'state_ml_id': state_ml_id})
        return state_id

    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = super().select_one_dict_by_column_and_value(
            column_value=location_id,
            select_clause_value=LocationLocalConstants.STATE_TABLE_COLUMNS)
        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_dict=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    def get_state_id_by_state_name(self, state_name: str,
                                   lang_code: LangCode = LocationLocalConstants.DEFAULT_LANG_CODE) -> int or None:
        logger.start("start get_state_id_by_state_name",
                     object={'state_name': state_name})
        if state_name is None:
            logger.end(log_message="State name was not provided",
                       object={'state_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or LangCode.detect_lang_code(state_name)
        where_clause = f"state_name='{state_name}' AND lang_code='{lang_code.value}'"

        self.state_ml = self.state_ml or StateMl()
        state_id_dict = self.state_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.STATE_ID_COLUMN_NAME,
            where=where_clause,
            order_by="state_id DESC")
        state_id = state_id_dict.get(
            LocationLocalConstants.STATE_ID_COLUMN_NAME)

        logger.end("end get_state_id_by_state_name",
                   object={'state_ids': state_id})
        return state_id
